/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

//  xxxxxxxxxxxxx          version: 1.0              xxxxxxxxxxxxxxx
		 
//  xxxxxxxxxxxxxxx   global variable declarations   xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var ptype = '1';   // Paradigm type ('1' = all list are equal or '2' = all list have different strings)

var stype = '3';  	// Stimulus type. range: [1..10]. see html element for details or source code: stimGen(),..

var dtype = '2';  	// Stimulus display mode ('1' = all stimuli in the list at once, '2' = each stimulus in the list separately, '3' = cumulative)

var item_cnt = 3;  // Maximum number of lists (items) to display

var row_cnt = 3;   // Number of stimuli (rows, lines) in each stimulus list

var alfa_cnt = 4;  //  String length of each element in list. Number of alphanumeric characters in each row.

var ttype = '1';    // Task type - scoring type. range= '1' or '2'. see html ttype element

var font_family = 'Georgia, serif';  // default canvas ctx.fontFamily 

var i_time = 1000;    // Interstimulus time interval. Time before the first stimulus is displayed in the list (ms)

var p_time = 1000;    // The time between the display of individual stimuli in the list (ms)

var s_time = 1000;    // Stimulus display duration (ms)

var d_time = 1000;	  // Delay time (ms)

var timer_enabled = false;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var timer = 600;      // Task time in seconds if timer is enabled

var is_mis = true;    // if (is_mis==true) => missing values are allowed

var prg_type = true;   //  type of information in Progress Bar (true => List/Items, false => Nothing) 

var demo = true;     // if (demo == true) => Show task solutions 
  
var back_clr = '#fff';  // background canvas color during task

var pen_clr = '#000';  //  font color (base font) during the task 
	
var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 
 

// Other global constants and variables:

var even_odd = {};       // for each even or odd characters in string. used for stype range: [1..4].

var even_ar = [];        // for each even or odd characters in string. used for stype range: [1..4].

var user_list = [];     // textarea for User-defined stimuli. used if stype range: [9..10]

var is_storage = false; // local storage is enabled on user browser or not
   
var dirty = false;  // if  (dirty==true) => user changed any options in form 

var is_task = false;    // task in progress or not. is_task is true for all events between startTask() and theEnd() events

var startTime;      // timestamp (start time) for current stimulus

var item;   		// current item in task     

var is_prepared;     // next stimulus are prepared or not => for window.requestAnimationFrame() callbacks 

var c_string;      // current last string in list to be displayed 

var task_begin;  		// task start time

var rad;           // text radius. for ctx.fillText() calls

var err = false;  // if true => page resize or Print screen press during task

// global (window) arrays
	
var stim = [];   // stimulus array (string in list for each item) dimensions = [0..item_cnt-1] x [0..row_cnt-1].  1. index is list item/list display number.  2. index is row (string in list for this item/display) 

var ans = [];   //  the user's response to each list in the task. dim = item x number of non-empty lines in each answer 

var mis = [];   // for each item (list display): if true => there is no user entered values (answers) for this item (missing value)



// application init [check storage, show start panel..] 
function init() 
{   
  is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }			
 
	loadFont();
	setIns();

	document.getElementById('answer').onpaste= function(){return false}; // Prevent pasting in answer textarea element
	document.getElementById('even_odd').value= 'A\nE\nI\nO\nU\n';
  document.onkeyup = catchPS; 

	showStart(); 
}


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// fill font-family names from array into select html element
function loadFont()  
{
	// used font-family array: The name of a font family. For example, "Times" and "Helvetica" are font families. Font family names containing whitespace should be quoted
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp	
	const arr = ['"Times New Roman", Times, serif', 'Georgia, serif', '"Palatino Linotype", "Book Antiqua", Palatino, serif', 'Arial, Helvetica, sans-serif', '"Arial Black", Gadget, sans-serif', '"Comic Sans MS", cursive, sans-serif', 'Impact, Charcoal, sans-serif', '"Lucida Sans Unicode", "Lucida Grande", sans-serif', 'Tahoma, Geneva, sans-serif', '"Trebuchet MS", Helvetica, sans-serif', 'Verdana, Geneva, sans-serif', '"Courier New", Courier, monospace', '"Lucida Console", Monaco, monospace' ];
	let sel= document.getElementById('font_family'), len = arr.length, opt; 
	sel.innerHtML= ''; 
	for (let i= 0; i!= len; i++) 
	{  
		opt = document.createElement('option');
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= arr[i];   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);                 			
	} 
	sel.selectedIndex=1; // Georgia, serif
	document.getElementById('font_family_msg').style.fontFamily= sel.value; // = arr[1];
}	


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		   
		//  dtype = '1';  	// Stimulus display mode ('1' = all stimuli in the list at once, '2' = each stimulus in the list separately, '3' = cumulative)
	if ( dtype == '1') { 
		document.getElementById('ins_dtype').style.display = 'none';
	}
	else  {
		document.getElementById('ins_dtype').style.display = 'inline';
		document.getElementById('ins_dtype').textContent= dtype == '2' ? 'The stimuli will not be displayed at the same time, but always one by one (only the first, only the second, only the third,..).' : 'The stimuli will not be displayed at the same time, but one by one cumulatively, ie: first stimulus, then first + second stimulus, then first + second + third stimulus..';
	}
		// ttype = '1';    // Task type - scoring type. range= '1' or '2'. see html ttype element
	document.getElementById('ins_ttype').textContent=  ttype == '1' ? "You don’t have to remember the order in which the stimuli appear." : "You must also remember the order in which the stimuli appeared, ie you must enter the stimuli exactly in the order in which they appeared. Enter the first stimulus in the first line, enter the second stimulus in the second line,..";
		// ptype = '1';   // Paradigm type ('1' = all list are equal or '2' = all list have different strings)
	document.getElementById('ins_ptype').textContent=  ptype == '1' ? 'The same list of stimuli is always displayed. The task is completed when you have entered all the stimuli correctly.' :  'A list of different stimuli will be displayed each time. The task is complete when you get to the last list of stimuli.';
	document.getElementById('ins_demo').style.display = demo ? 'block' : 'none';
	document.getElementById('ins_timer').style.display = document.getElementById('timer_enabled').checked ? 'block' : 'none';
}	


// deprecated but still works: https://developer.mozilla.org/en-US/docs/Web/API/KeyboardEvent/keyCode
// https://developer.mozilla.org/en-US/docs/Web/API/KeyboardEvent/code
// https://caniuse.com/#search=KeyboardEvent.keyCode
// https://caniuse.com/#search=KeyboardEvent.code
// This function is implemented because of possible cheating (we can not prevent the use of cameras on mobile phones)
function catchPS(e)   
{ 
  e = e || window.e;  
  if (is_task && e.key === 'PrintScreen') {  
    err=true;	  // Print screen error 
  } 
}


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// raised when user change font-family on html <select> element
function tryFont(e)  {	
	document.getElementById('font_family_msg').style.fontFamily= e.value;
}	


// raised when page is resized. see html: <body onresize="pageResize()"> => stretch canvas 
function pageResize()  
{   
	if (is_task)  { 
		err=true;	  // resize error										 
	}	
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty=true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{

   case 'ptype':                                                                                                                                           
			msg.textContent= (e.value === '1') ? 'The same list of stimuli is always displayed. The task is complete when all strings within the list have been entered correctly (either the maximum number of list views has been reached or the Timer has stopped the task).' : 'A different list of stimuli is always displayed. The task is completed when the maximum number of list views has been reached (or the Timer has stopped the task).';
			break;


		case 'stype':                                                                                                                                           
			/*  Stimulus type: stype var range [1..10]
					'1' => Even-odd lowercase
					'2' => Odd-even lowercase
					'3' => Even-odd UPPERCASE (default)
					'4' => Odd-even UPPERCASE
					'5' => Digits
					'6' => Letters-lowercase
					'7' => Letters-UPPERCASE
					'8' => All digits and letters
					'9' => User defined list - lowercase
          '10' => User defined list- UPPERCASE */
			msg.textContent= e.options[e.selectedIndex].getAttribute('data-help');	
			if ( e.value == '1' || e.value == '2' || e.value == '3' || e.value == '4' )  
			{
				document.getElementById('even_odd_fs').style.display= 'block';
				document.getElementById('even_odd_l').textContent= ( e.value == '1' || e.value == '3') ? 'Define the characters in each EVEN position (2,4,6,..) in the string:' : 'Define the characters in each ODD position (1,3,5,..) in the string:';
				document.getElementById('even_odd_msg').textContent= ( e.value == '1' || e.value == '3' ) ? 'Each character in this list will randomly appear at each EVEN position (2,4,6,..) in the string. Eg if you enter vowels: A, E, I, O, U. Then, for example, these random stimuli will be generated (depending on whether you choose uppercase or lowercase letters) eg: JAGE, MUTA, POLI,.. or eg: zuga, olaf, pajo..' : 'Each character in this list will randomly appear at each ODD position (1,3,5,..) in the string. Eg if you enter vowels: a, e, i, o, u. Then, for example, these random stimuli will be generated (depending on whether you chose uppercase or lowercase letters) eg: ugit, atif, ezur,.. or eg: AJEH, UKIZ, OLAP..';	
			  if ( e.value == '1' || e.value == '2'  )  {
					document.getElementById('even_odd').setAttribute('style','text-transform:lowercase');			
				}
				else /*if ( e.value == '3' || e.value == '4'  ) */ {   
					document.getElementById('even_odd').setAttribute('style','text-transform:uppercase');		
				}
			}
			else  {
				document.getElementById('even_odd_fs').style.display= 'none';
			}
			document.getElementById('user_list_fs').style.display= e.value == '9' || e.value == '10' ? 'block' : 'none';
			if (e.value == '9')  {
				document.getElementById('user_list').setAttribute('style','text-transform:lowercase');	
				document.getElementById('user_list_l').textContent= 'User-defined stimuli (lowercase). Define your own stimuli within the list:';			
			}
			else if (e.value == '10') {
				document.getElementById('user_list').setAttribute('style','text-transform:uppercase');
        document.getElementById('user_list_l').textContent= 'User-defined stimuli (UPPERCASE). Define your own stimuli within the list:';				
			}
			document.getElementById('p_parad').style.display= e.value == '9' || e.value == '10' ? 'none' : 'block';	
			break;
    
		//case 'dtype': break;	
    //case 'ttype': break;
    //case 'font_family': break;

		case 'item_cnt':
    case 'row_cnt':		
		case 'alfa_cnt': 		
		case 's_time':  		   
		case 'd_time': 		
			if (e.validity.valid)  {
				msg.textContent= ''; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 			
			break;  
		
		case 'i_time':
			if (e.validity.valid)  {
				msg.textContent= 'This time is approximate (strongly depends on the performance of the browser or OS).'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 			
      break;
				
		case 'p_time':
			if (e.validity.valid)  {
				msg.textContent= 'Stimuli are prepared during this period. So leave this period longer, especially for very short stimulus displays.'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 			
      break;
		
    case 'timer_enabled':   
      if (e.checked)  {
        msg.textContent= 'Yes. If the time has elapsed, the task will be interrupted by clicking on the button to display the next list.';
        document.getElementById('p_timer').style.display= 'block';
      }
      else  {  
        msg.textContent= 'No';
        document.getElementById('p_timer').style.display= 'none'; 
      }       
      break;
		
    case 'timer':
			if (e.validity.valid)  {
				msg.textContent= stringTime(e.value); 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break;
    
    case 'is_mis': 
    	msg.textContent= e.checked ? 'Missing values are allowed. The user does not have to enter a value.' : 'Missing values are not allowed. To move to the next list of stimuli, any non-empty value must be entered.';    
			break;
			    
    case 'prg_type':                                                                                                                                                   
      msg.textContent= e.checked ? 'Yes' : 'No';     
      break;    
				
		case 'demo':
			msg.textContent= e.checked ? 'The correct solution will be displayed.' : 'No';  
			break; 
		
		case 'back_clr': 		// background canvas color during task
		case 'pen_clr':  		//  font color (base font) during the task   
			msg.textContent= ( validColor(e.value) ) ? '' : 'The selected color is invalid.'; 
			break; 	
		
		case 'store':                                                                                                                                           
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty == true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity.valid object in all input type="number" + check colors
function optionsValid()  
{      
	let res= document.getElementById('item_cnt').validity.valid && 
	document.getElementById('row_cnt').validity.valid &&	
	document.getElementById('alfa_cnt').validity.valid && 		 
	document.getElementById('i_time').validity.valid &&		
	document.getElementById('p_time').validity.valid &&	
	document.getElementById('s_time').validity.valid &&
	document.getElementById('d_time').validity.valid &&	
	document.getElementById('timer').validity.valid &&		
	validColor(document.getElementById('back_clr').value) && 
	validColor(document.getElementById('pen_clr').value);
  return res;              
}


// start task (research) 
function startTask(e)   
{     
  if ( !validateTextarea() ) {
		return;
	}

  initLayout();  

	setPar(); 
	
	stimGen();
 
  setFont(); 
 
  playItem();	     
}


// init page layout when task is started. init visible/invisible html dom elements, progress bar
function initLayout()
{

	setCanvas();	
	
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
 	document.getElementById('can').style.visibility = 'hidden';
	//document.getElementById('answer_section').style.display= 'block'; 
	document.getElementById('nav_items').textContent= '';	
  if (!prg_type)  {
		document.getElementById('progress_holder').style.display= 'none';    	
	}	
	else  {
		document.getElementById('progress_holder').style.display= 'block';   				   			
		document.getElementById('progress').setAttribute('style', 'width:100%');  // reset progress bar width						 	
	}
  
	/*  Stimulus type: stype var range [1..10]
	'1' => Even-odd lowercase
	'2' => Odd-even lowercase
	'3' => Even-odd UPPERCASE (default)
	'4' => Odd-even UPPERCASE
	'5' => Digits
	'6' => Letters-lowercase
	'7' => Letters-UPPERCASE
	'8' => All digits and letters
	'9' => User defined list-lowercase 
  '10' => User defined list-UPPERCASE */
	
	switch (stype)  {
			case '1':
			case '2':
			case '6':
			case '9':
			  document.getElementById('answer').setAttribute('style','text-transform:lowercase');	
				break;
				
			case '3':
      case '4':
			case '7':
			case '10':
			  document.getElementById('answer').setAttribute('style','text-transform:uppercase');	
				break;
			
			default:		// '5' => Digits. '8' => All digits and letters	
			  document.getElementById('answer').setAttribute('style','text-transform:none');
        // document.getElementById('answer').removeAttribute('style');					
		}
			
}


// center stimulus (canvas) on user screen (approx)
function setCanvas()  
{
	let c= document.getElementById('can'), ctx= c.getContext('2d');	 
	c.width= innerWidth;  
	c.style.width= c.width + 'px';	     
	c.height= innerHeight;   
	c.style.height= c.height + 'px';  

  ctx.clearRect(0, 0, c.width, c.height);  
	//c.style.display= 'block';
	
	ctx.textAlign = 'center';
	ctx.textBaseline= 'middle';	
	
	ctx.fillStyle= back_clr;
	ctx.strokeStyle= pen_clr; 	

}


/*  Stimulus type: stype var range [1..10]
	'1' => Even-odd lowercase
	'2' => Odd-even lowercase
	'3' => Even-odd UPPERCASE (default)
	'4' => Odd-even UPPERCASE
	'5' => Digits
	'6' => Letters-lowercase
	'7' => Letters-UPPERCASE
	'8' => All digits and letters
	'9' => User defined list-lowercase 
  '10' => User defined list-UPPERCASE */
// set canvas font for latter draw text => calls ctx.fillText()
function setFont()  {
  let c= document.getElementById('can'), ctx= c.getContext('2d'), txt= '', txt_w;		
	//ctx.font= '10px Arial, Verdana, Helvetica, sans-serif';   // https://www.w3schools.com/cssref/css_websafe_fonts.asp

  let width = c.width,  height = c.height / row_cnt;  // cell width, cell height
  let r= Math.floor( Math.min(width,height)/2 ) + 1;  // default (tested) font radius
	
	if ( stype == '9' || stype == '10') {
    for (let row in stim[0])  { // all items i have equal strings in each row => stim[i][row]= const;
			if ( stim[0][row].length > txt.length )  {
				txt= stim[0][row];
			}
		}
	}
  else  {
    txt= stim[0][0];  // fixed row_cnt in each item
	}		
  
	do  { 
	  r--;
		ctx.font = r.toString() + 'px ' + font_family; 
		txt_w = ctx.measureText(txt).width;   
	} while (txt_w >= width);
	
	// https://stackoverflow.com/questions/1134586/how-can-you-find-the-height-of-text-on-an-html-canvas/45789011
	let metrics = ctx.measureText(txt), tmp;
	let fontHeight = metrics.fontBoundingBoxAscent + metrics.fontBoundingBoxDescent;  // Firefox not implemented yet.. see: https://caniuse.com/
	let actualHeight = metrics.actualBoundingBoxAscent + metrics.actualBoundingBoxDescent;
	if (fontHeight && actualHeight) {
    tmp= Math.min(fontHeight, actualHeight);
	}		
	else if (fontHeight)  {
    tmp= fontHeight;
	}
	else if (actualHeight)  {
    tmp= actualHeight;
	}	
  else  {
    tmp= r;  // IE11 => see: https://caniuse.com/
	}		
	rad= Math.floor(Math.min(r,tmp));
	ctx.font = rad.toString() + 'px ' + font_family; 
}


// set internal application research parameters
// reset all global vars and arrays
function setPar()  
{	
	if ( stype == '9' || stype == '10' )  {  // User defined list: 9 or 10
    row_cnt= user_list.length;
		ptype= '1';  // Paradigm type pytpe= '1' => all items have equal list
		document.getElementById('ptype').selectedIndex= 0;
	}	
	
	// init all global arrays
	ans= []; 	 //  stim = []; 	=> stimGen()
	mis=[];
	for (let i= 0; i< item_cnt; i++)   {  // i => item        
    ans[i]= [];
		mis[i]= true;
  } 
	
	// reset all global vars
  item = 0;  	
	is_task= true;     //  see pageResize()
  err= false;	
	timer_enabled= (document.getElementById('timer_enabled').checked) ? true: false;  // timer is enabled or disabled
	task_begin = new Date();
}


// stimulus generator. generates task stimuli 
function stimGen()	
{  
  stim = []; 	// stim array dimensions are = [0..item_cnt-1] x [0..row_cnt-1] 
	for (let i= 0; i < item_cnt; i++)  
	{		
		if ( i == 0 ) {
			genItem(0);
		}
		else if ( ptype == '1' || stype == '9' || stype == '10' )  {    		
      stim[i] = stim[0];			
		}
		else {
			genItem(i);
		}
	}
}


/* generate new item. i => item number. range= [0..item_cnt-1]
   each string in each row must be different => all elements in list are different 
	 new stimulus will be different from the previous one in all rows  
  Stimulus type: stype var range [1..10]
	'1' => Even-odd lowercase
	'2' => Odd-even lowercase
	'3' => Even-odd UPPERCASE (default)
	'4' => Odd-even UPPERCASE
	'5' => Digits
	'6' => Letters-lowercase
	'7' => Letters-UPPERCASE
	'8' => All digits and letters
	'9' => User defined list-lowercase 
  '10' => User defined list-UPPERCASE */
function genItem(i) 
{	
	stim[i]= [];    	// stim array dimensions are = [0..item_cnt-1] x [0..row_cnt-1]
	// sel = object for check if this string already exists in list
	// s = characters in each row (0..row_cnt-1) => stimulus in each row 	
	let sel = {}, s, rnd, user, rnd_last, user_last;
	for (let row = 0; row < row_cnt; row++)
	{
		do 
		{    
			s = '';   			
			switch (stype)
			{					
				case '1':  // '1' => Even-odd lowercase 
				case '3':  // '3' => Even-odd UPPERCASE					 
					for (let i = 0; i < alfa_cnt; i++)  
					{
						if ( i % 2 == 0 )  
						{
							do 	{
								rnd= stype == '1' ? lowerStrGen(1) : upperStrGen(1);
							}  while ( typeof  even_odd[rnd] !== 'undefined' || rnd == rnd_last); 
							s+= rnd;
							rnd_last= rnd;
						}
						else  
						{
							do  {
								user= rndFromArray(even_ar); 
							} while ( user == user_last);
							s+= user;
							user_last= user;
						}
					}
					break;
				
				case '2':  // 	'2' => Odd-even lowercase
				case '4':  //   '4' => Odd-even UPPERCASE				
						
					for (let i = 0; i < alfa_cnt; i++)  
					{
						if ( i % 2 == 0 )  
						{
							do  {
								user= rndFromArray(even_ar); 
							} while ( user == user_last);
							s+= user;
							user_last= user;			
						}
						else  
						{
							do 	{
								rnd= stype == '2' ? lowerStrGen(1) : upperStrGen(1);
							}  while ( typeof  even_odd[rnd] !== 'undefined' || rnd == rnd_last); 
							s+= rnd;
							rnd_last= rnd;
						}
					}
					break;
					
				case '5':  // '5' => Digits 		
					let last, d;
					for (let i=0; i<alfa_cnt; i++)  {
						do  {
							d= getRndInt(0,9).toString();
						}	while ( d == last );   // each digit must be different from previous
						last = d;
						s+= d;
					}			  
					break;
					
				case '6':  // 	'6' => Letters-lowercase
					s= lowerStrGen(alfa_cnt);			
					break;
					
				case '7':  // '7' => Letters-UPPERCASE
					s= upperStrGen(alfa_cnt);			
					break;
					
				case '8':  // 	'8' => All digits and letters 
					s= allStrGen(alfa_cnt);
					break;
				
				case '9':  	//	'9' => User defined list-lowercase 
				case '10':  //  '10' => User defined list-UPPERCASE 
					s= user_list[row];
					break;
				
			}  // switch
		} while ( typeof(sel[s]) !== 'undefined' ); // new stimulus must be different from the all previous stimuli
		sel[s]= 1;
		stim[i][row]= s;
	}  // for row	
}	


// items = 1.dim array
function rndFromArray(items) {
  return items[Math.floor(Math.random()*items.length)];  
}


// generates a random string from lowercase letters
// each generated character must be different
function lowerStrGen(cnt)
{ 
  let all = "abcdefghijklmnopqrstuvwxyz";
  let txt = "", up= all.length, key= {}, c;
  for (let i = 0; i < cnt; i++)  {
		do  {
			c= all.charAt(Math.floor(Math.random() * up));
		}	while ( key[c] == 1 ); 
		key[c]= 1;
    txt += c;
	}
  return txt;
}


// generates a random string from uppercase letters
// each generated character must be different 
function upperStrGen(cnt)
{ 
  let all = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  let txt = "", up= all.length, key= {}, c;
  for (let i = 0; i < cnt; i++)  {
    do  {
			c= all.charAt(Math.floor(Math.random() * up));
		}	while ( key[c] == 1 ); 
		key[c]= 1;
		txt += c;
	}
  return txt;
}


// generates a random string from all (uppercase and lowercase) letters and digits
// each generated character must be different 
function allStrGen(cnt)
{ 
  let all = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
  let txt = "", up= all.length, key= {}, c;
  for (let i = 0; i < cnt; i++)  {
    do  {
			c= all.charAt(Math.floor(Math.random() * up));
		}	while ( key[c] == 1 ); 
		key[c]= 1;
		txt += c;
	}
  return txt;
}


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   

  if ( document.getElementById('ptype').selectedIndex > - 1 )  {
		ptype= document.getElementById('ptype').value;
	}

	if ( document.getElementById('stype').selectedIndex > - 1 )  {
		stype= document.getElementById('stype').value;
	}
  
	if ( document.getElementById('dtype').selectedIndex > - 1 )  {
		dtype= document.getElementById('dtype').value;
	}
	
	if ( document.getElementById('item_cnt').validity.valid )  {  
		item_cnt= parseInt(document.getElementById('item_cnt').value, 10);			
  }
	
	if ( document.getElementById('row_cnt').validity.valid )  {  
		row_cnt= parseInt(document.getElementById('row_cnt').value, 10);			
  }

  if ( document.getElementById('alfa_cnt').validity.valid )  {  
		alfa_cnt= parseInt(document.getElementById('alfa_cnt').value, 10);			
  } 

  if ( document.getElementById('ttype').selectedIndex > - 1 )  {
		ttype= document.getElementById('ttype').value;
	}

	font_family= document.getElementById('font_family').value;

	if (document.getElementById('i_time').validity.valid) {
		i_time= parseInt(document.getElementById('i_time').value, 10);
  }

  if (document.getElementById('p_time').validity.valid) {
		p_time= parseInt(document.getElementById('p_time').value, 10);
  }

	if ( document.getElementById('s_time').validity.valid )  {  
		s_time= parseInt(document.getElementById('s_time').value, 10);			
  }
	  
	if (document.getElementById('d_time').validity.valid) {
		d_time= parseInt(document.getElementById('d_time').value, 10);  
  }
  
	// timer_enabled => update latter in setPar()
	
	if (document.getElementById('timer').validity.valid) {
		timer= Math.round(parseInt(document.getElementById('timer').value, 10));  // task time (s)
  }	
	
  is_mis= document.getElementById('is_mis').checked ? true: false; 
	
	prg_type= document.getElementById('prg_type').checked ? true : false;  // display item count on footer or nothing
		
	demo= document.getElementById('demo').checked ? true: false;	
		
	if (validColor(document.getElementById('back_clr').value))  {		
		back_clr= document.getElementById('back_clr').value;
  }
	
	if (validColor(document.getElementById('pen_clr').value))  {		
		pen_clr= document.getElementById('pen_clr').value;
  }
		
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;
}


// Validates and filters values entered in 2 lists (textarea)
// return true if Even-Odd list is ok or User defined list is ok
function validateTextarea()  
{	
  /* Stimulus type: stype var range [1..10]
	'1' => Even-odd lowercase
	'2' => Odd-even lowercase
	'3' => Even-odd UPPERCASE (default)
	'4' => Odd-even UPPERCASE
	'5' => Digits
	'6' => Letters-lowercase
	'7' => Letters-UPPERCASE
	'8' => All digits and letters
	'9' => User defined list-lowercase 
  '10' => User defined list-UPPERCASE */
  
	even_odd= {};   // init global window object even_odd
	even_ar= [];   // init global window array even_odd
	user_list= [];  // init global window array user_list
		
	let lines, v, sel = {}, len;
	let result= true;  // default for all other stype
	if ( stype == '1' || stype == '2' || stype == '3' || stype == '4' )  
	{
		// Enter one character in each line here. The number of different characters you can enter is 2-12. Duplicates will be deleted. You may only enter one character per line (the rest is ignored).
		lines = document.getElementById('even_odd').value.split('\n');
		for (let i = 0; i < lines.length; i++)
		{
			v= trim(lines[i]);  // code here using lines[i] which will give you each line			
			if ( v.length > 0 )  
			{
				if ( stype == '1' || stype == '2') {  // '1' => Even-odd lowercase. '2' => Odd-even lowercase 
					v= v.toLocaleLowerCase();
				}
				else  {  	// '3' => Even-odd UPPERCASE. '4' => Odd-even UPPERCASE
					v= v.toLocaleUpperCase();
				}				
				v = v[0];  // only 1. char from string
				if ( sel[v] != 1   )  {  //  if (typeof(v) !== 'undefined')
					even_ar.push(v);
					sel[v] = 1;
				}
			}
		}
		len = even_ar.length;
		if ( len > 0 )  
		{
			// init global window object even_odd..
			for (let key = 0; key < len; key++)  {
				even_odd [ even_ar[key] ] = even_ar[key]; 
			}		
			document.getElementById('even_odd').value= even_ar.join('\n');
		}
		else  {
			document.getElementById('even_odd').value= '';
		}
		if ( len < 2 || len > 12 ) {
      result= false;
			let e= document.getElementById('is_valid');
			e.textContent='Even-Odd list must have 2..12 valid characters (rows).';   // These settings are marked in red. 
			e.setAttribute('style', 'text-align:center;color:#f00;font-weight:bold'); 
		}			
	}
	
	else if ( stype == '9' || stype == '10' )  
	{
		//Enter one word (or any string) in each line. The range of words you are allowed to type is 1..50. Duplicates are deleted. Maximum allowed length of each word (string) is 30.
		lines = document.getElementById('user_list').value.split('\n');
		for (let i = 0; i < lines.length; i++)
		{
			v= trim(lines[i]);  // code here using lines[i] which will give you each line
			if ( v.length > 0 ) 
			{
				v = v.substring(0, 30);   // First character is at index 0.			
				if ( stype == '9' ) {  // '9' => User defined list-lowercase 
					v= v.toLocaleLowerCase();
				}
				else  {  	// '10' => User defined list-UPPERCASE
					v= v.toLocaleUpperCase();
				}
				if ( sel[v] != 1   )  {  //  if (typeof(v) !== 'undefined')
					user_list.push(v);
					sel[v] = 1;
				}
			}
			if ( i > 49 ) {
				break;
			}
		}
		len = user_list.length;	
		if ( len > 0 )  {
			document.getElementById('user_list').value= user_list.join('\n');
		}
		else  {
			document.getElementById('user_list').value= '';
		}
		if ( len < 1 )  {
      result= false;
			let e= document.getElementById('is_valid');
			e.textContent='User defined list is empty.';   // These settings are marked in red. 
			e.setAttribute('style', 'text-align:center;color:#f00;font-weight:bold');  
		}			
	}	

	return result;
}	


// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  let hh = Math.floor(total/(60*60));         
  let div_min = total % (60 * 60);
  let mm = Math.floor(div_min / 60);       
  let div_sec = div_min % 60;
  let ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


//  play (display) next list or end of task
function nextItem()  
{ 	
  if (err)  {
		theEnd(true, null);  // page resize error or Print screen error
		return;
	}
  
	if (timer_enabled)  {	// check time
		let endDate= new Date();   //tt= stringTime(tt);
		let tt= Math.round( (endDate.getTime() - task_begin.getTime() ) / 1000);  // task time 
		if ( tt > timer )  {
			theEnd(false, item_cnt-1);
			return;
		}
	}

	/* store user answers (lines) for this item
  Stimulus type: stype var range [1..10]
	'1' => Even-odd lowercase
	'2' => Odd-even lowercase
	'3' => Even-odd UPPERCASE (default)
	'4' => Odd-even UPPERCASE
	'5' => Digits
	'6' => Letters-lowercase
	'7' => Letters-UPPERCASE
	'8' => All digits and letters
	'9' => User defined list-lowercase 
	'10' => User defined list-UPPERCASE */
		
	let lines, v, arr= [];
	lines = document.getElementById('answer').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		v = trim(lines[i]);  
		switch (stype)  {
			case '1':
			case '2':
			case '6':
			case '9':
			  v= v.toLocaleLowerCase();
				break;
				
			case '3':
      case '4':
			case '7':
			case '10':
			  v= v.toLocaleUpperCase();
				break;
			
			// default:		// '5' => Digits. '8' => All digits and letters	
			  // do nothing	
		}
		if ( v.length > 0 )  { 
				arr.push(v);
		}
	}	
	
	let len = arr.length;
	if ( len > 0 )  {
		document.getElementById('answer').value= arr.join('\n');
		ans[item]= arr;   // trimed + converted to lowercase (uppercase) if needed 
	}
	else  {
		document.getElementById('answer').value= '';
		//ans[item]= null;
	}
	mis[item]= len == 0 ? true : false;
	if ( !is_mis && len == 0 ) {   // missing values are not allowed
	  alert('You must enter any non-blank answer.'); // 'You must enter any non-blank answer. Try to guess.
		return;
	}
	
	let points = scoreItem(item).length;
	// check if task is done. all elements in list are entered ok
	if ( points == row_cnt && ptype == '1')  {
		theEnd(false, item);
    return;  
  }	
	
	// play next list display or end of task
	item++;					
	if (item < item_cnt)  {         		 
    playItem();
  }
  else  {               
    theEnd(false, item_cnt-1);  // end of task. no resize error in this task 
  }
}


// trim input string (user answer)
function trim(str) {
   return str.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
}
 
 
// show next item => start callbacks after TimeOut delay 
function playItem()
{
	document.getElementById('answer_section').style.display= 'none';
	is_prepared = false;

		//setTimeout(prepare, i_time);  => don't do this!
		//setTimeout(startCallbacks, i_time);   // this is ok
	setTimeout("window.requestAnimationFrame(prepare);", i_time);   // this time period is not precize
	
	c_string =  dtype == '1' ? stim[item].length - 1 : 0;   		
}

/*
function startCallbacks() {
	window.requestAnimationFrame(prepare);  
}*/	

	
// browser calls this function to prepare an animation before the next repaint (callback)
// this is 1. callback. main function: prepare stimulus in period before stimulus
// prepare next stimulus
function prepare(timestamp) 
{ 
	if (!is_prepared)  {   		
		startTime = timestamp;	// reset time when the stimulus sequence started 
    if (dtype == '1') {
			drawAll();			
		}
		else {
			drawPartial();
		}
		is_prepared= true;		
		//document.getElementById('can').style.visibility = 'hidden';		
	}	
	if ( timestamp - startTime <= p_time )  {  // just wait //let elapsed = timestamp - startTime; 	
		window.requestAnimationFrame(prepare);   // Note: requestAnimationFrame() is 1 shot.
	}
	else  { 		  			
		startTime = timestamp;	// timestamp when the alphanumeric characters display started 
		document.getElementById('can').style.visibility = 'visible';  // only once
		window.requestAnimationFrame(displayString);  // set another callback => dispatch
	}		
}


// draw all stimulus in list on hidden canvas (prepare and wait) 
function drawAll() 
{
	let c = document.getElementById('can'), ctx = c.getContext('2d'), width = c.width, height = c.height;					
	ctx.fillStyle= back_clr;
	ctx.fillRect(0, 0, width, height);    // clearRect() method of the Canvas 2D API erases the pixels in a rectangular area by setting them to transparent black. 
  ctx.fillStyle= pen_clr;
	height = height / row_cnt;
  let	top, xs, ys, txt;  
	for (let row = 0; row < row_cnt; row++)  {	   								
		top = row * height;  // cell top bound
		xs = Math.round(width/2);
		ys= Math.floor(top+height/2);  			      
		txt= stim[item][row]; 
		ctx.fillText(txt, Math.round(width/2), Math.round(top+height/2));
	} 
}	


// draw some stimulus in list on hidden canvas (prepare and wait) 
function drawPartial() 
{
	let c = document.getElementById('can'), ctx = c.getContext('2d'), width = c.width, height = c.height, top, r, txt;					
	ctx.fillStyle= back_clr;
	ctx.fillRect(0, 0, width, height);  // clearRect() method of the Canvas 2D API erases the pixels in a rectangular area by setting them to transparent black.
  ctx.fillStyle= pen_clr;
  if (dtype == '3') {  // dtype == '3' => cumulative (1, 1 + 2, 1 + 2 + 3, ..)
		height = height / row_cnt;  // decrease height:  height= height  / ( c_string + 1);  // cell height
		for (let row = 0; row <= c_string; row++)  
		{	   								
			top = row * height;  // cell top bound  			      
			txt= stim[item][row]; 
			ctx.fillText(txt, Math.round(width/2), Math.round(top+height/2));
		}
  }
	else  {  // dtype == '2' => each stimulus in the list separately
			txt= stim[item][c_string]; 
			ctx.fillText(txt, Math.round(width/2), Math.round(height/2));
	} 
}	


// handle stimulus duration.  
// trigger end of display of stimulus for all Stimulus display mode (dytpe)  
function displayString(timestamp)  {
	//let elapsed = timestamp - startTime; 
	if ( timestamp - startTime < s_time ) {  //   var s_time => alphanumeric characters duration (ms)
		window.requestAnimationFrame(displayString);  // wait
	}
	else  {   			
		document.getElementById('can').style.visibility = 'hidden';  // only once  
		startTime = timestamp;	// timestamp when the stimulus display ended
		c_string++;
		if (dtype == '1' || c_string > stim[item].length -1)  {		
			window.requestAnimationFrame(displayDelay);  // set last callback
		}
		else  {
			is_prepared= false;
			window.requestAnimationFrame(prepare);  // call again 1. callback. not all strings (stimuli) in list are displayed yet 
		}
	}
}


// last callback. handle delay period between end of all stimulus in list and begining od display the Answer PAnel
function displayDelay(timestamp)  {
	if ( timestamp - startTime < d_time )  {   
		window.requestAnimationFrame(displayDelay);  // wait
	}
	else  {   	 	
	  showItem();
	}
}


// return random integer in range [min..max]
function getRndInt(min, max) {
		return Math.floor(Math.random() * (max - min + 1) ) + min;
}


// show current answer_section panel (textarea element + footer)    
function showItem() 
{   		   
	document.getElementById('answer_section').style.display= 'block';  
	
	let a= document.getElementById('answer'); 		
	//a.rows= row_cnt;
	a.value= '';	// always delete previous answers
	a.focus();		
	
	if (demo)  {			
		let len= stim[item].length, s= '';
		for (let row = 0; row < len; row++)  {
			s+= stim[item][row] + '\n';
		}	
		a.value= s;	
	}      
	
	updateFooter(); 
}


// update footer (navigation button, progress bar,..). items are in range: 0..item_cnt-1  	
function updateFooter() 
{			 
	let btn= document.getElementById('next');     
	if(item == stim.length-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update "Next item" image on "right arrow" button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Next';
	} 
	if (prg_type)  {    // update Progress bar => show current item count and remaining item count on Progress Bar (default)                                                                                                                                                       	
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent= (item + 1) + ' / ' + item_cnt;   // item/total
	}
}	


// return array (list) of correct entered values
// i = item number
function scoreItem(i)  {
	let up= stim[i].length, last = ans[i].length, res= {}, key, result= [];
	if ( ttype == '1' )
	{   // string positions within List and answer positions are NOT important
		for (let row = 0; row < up; row++)  
		{ // iterate within stimulus list in this item
			for (let j = 0; j < last; j++)  
			{  // iterate within user answers (id="answer" textarea filtered lines). see nextItem() 
				// this is redundant: ans[i][j].toLocaleLowerCase() because nextItem() is always done it before this function
				key= stim[i][row];
				if ( ans[i][j] == stim[i][row] && typeof res[key] === 'undefined')  {  		  
					res[key]= key; 
				}
			}
		}	
	}
	else  
	{  // string positions within List and answer positions ARE important
		for (let row = 0; row < up; row++)  
		{   // iterate within stimulus list in this item
				// this is redundant: .toLocaleLowerCase(). see nextItem(),  genItem(), validateTextarea() 
				if ( typeof ans[i][row] !== 'undefined' && ans[i][row] == stim[i][row] )  {
					key= stim[i][row];		// duplicates are eliminated in functions validateTextarea(), genItem()		 		 
					res[key]= key;	
			}
		}				
	}
  // transform object into array
  for (let k in res)  {
    result.push(res[k]);
	}		
	return result;
}	


/* end of task (research)
   if (is_error== true) => page (window) was resized during task (when is_task==true). Only show html error warning
   last => item (list) number  */
function theEnd(is_error, last)  
{     
	is_task= false;    
	
	document.getElementById('answer_section').style.display= 'none';	
  document.getElementById('can').style.visibility = 'hidden';
	
	document.getElementById('end').style.display= 'block';   
 
	if (is_error) {
		document.getElementById('error').style.display= 'block';
		document.getElementById('result_wrapper').style.display= 'none';
		document.getElementById('score').textContent= '';
		return;
	}
	
  document.getElementById('error').style.display= 'none';
	document.getElementById('result_wrapper').style.display= 'block';
	
	let blank= 0, sum= 0, up= stim[0].length-1, max_point=0, all_p= []; header = ['List', 'All values', 'Correct values', 'Points', 'Max-points'], rows = []; 
	for (let i = 0; i <= last; i++)    // i => item (list) number	
	{	
		let s= '';  // displayed stimulus correct content
		for (let j= 0; j < up; j++)  {  // not last one
			s+= stim[i][j] + '<br>';
		}
		s+= stim[i][up]; // last one without <br> 
		if ( mis[i] )  {  
		  blank++;		// blank = number of missing items in task 
		  rows[i]= [ parseInt(i+1,10), s, '',  0, row_cnt ];
		}
		else 	
		{
			let res = scoreItem(i);  // list of correct user-entered values
			let point = res.length;  // number of points for this item (displayed list)
			sum+= point;
			all_p.push(point);  // number of points for each stimulus list display
			let a= '';  // user-entered correct string values
      if (point > 0) 
			{
				let len = point-1;   
				for (let j= 0; j< len; j++) {  // not last one
					a+= res[j] + '<br>';
				}		
				a+= res[len];  // last one without <br> 			
			}
			rows[i]=  [ parseInt(i+1,10), s, a,  point, row_cnt ];
		}
		max_point+= row_cnt;
	}		 
	
	rows.push( ['Sum', '', '', sum, max_point ] );		
	
	createTableArray(document.getElementById('table1'), rows, header);
	
  let percent = (100 * sum / max_point).toFixed(3);
	
	document.getElementById('score').textContent= 'Points = ' + sum + '. Maximum points = ' + max_point + ' Percent = ' + percent + '%.'; 
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
	
	try 
  {
    if (store) {
			setStorage(sum, all_p.join(' '), blank);  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }  
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:	 
	 sum = sum of all points in task
	 all_points = number of points in each list diplayed 
	 blank = number of missing items in task */
function setStorage(sum, all_points, blank)   
{   
	let oldItems = JSON.parse(localStorage.getItem('vlsl-v1.0')) || []; 
  let id= oldItems.length+1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		        
			'points': sum, 
			'missing' : blank,
		  'all-points': all_points,			
			'paradigm' : ptype,
			'stimulus': stype,			
			'display': dtype,
			'max-list': item_cnt,
			'row-count': row_cnt,			
			'alpha-count' : alfa_cnt,
			'scoring' : ttype,			
			'inter-time': i_time,
			'between-time': p_time,			
			'display-time' : s_time,				
			'delay-time' : d_time,					
			'timer-time' : document.getElementById('timer_enabled').checked ? stringTime(timer) : '',
			'is-missing' : is_mis ? 1 : 0,  			      
			'demo': demo ? 1 : 0		
  };

  oldItems.push(newItem);

  localStorage.setItem('vlsl-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems);  
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined)	{
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').setAttribute('class', 'center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';        //.removeAttribute("disabled");  
  document.getElementById('hide_option').style.display = 'none';   
  document.getElementById('instructions').style.display= 'block';
  refMsg(); 
  setIns();	
	//validateTextarea();
}

// begin of new code for export localStorage app data:  data= localStorage.getItem('myApp-v1.0');

// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (document.getElementById('store_group').disabled === false && localStorage.getItem('vlsl-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('vlsl-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';   // xxx
		}
	}
  dirty= false;
  document.getElementById('is_valid').textContent= '';  // tmp reset validator message	
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('vlsl-v1.0') ? JSON.parse(localStorage.getItem('vlsl-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    //alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved application results: '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 {  
   window.localStorage.removeItem('vlsl-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('vlsl-v1.0');
	if (!data)  {
    //alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('vlsl-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved cumulative data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// displays Latest research results in new empty web page (tab) 
function newWinLast()
{ 
	let e = document.getElementById('table1');	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{ 
	let	tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	} 
}	


// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
